@@ -10,29 +10,32 @@ import json
 import os
 import pathlib
 import re
+import socket
 import stat
 import struct
 import subprocess
 import tempfile
+import textwrap
+import threading
 import time
 import typing
 import warnings
 from contextlib import contextmanager
+from dataclasses import asdict
 from typing import Optional, Union
 
 import apkutils2
 import requests
 from deprecation import deprecated
+from PIL import Image, UnidentifiedImageError
 from retry import retry
 
-from ._adb import BaseClient
+from ._adb import AdbConnection, BaseClient
 from ._proto import *
 from ._utils import (APKReader, ReadProgress, adb_path, get_free_port,
                      humanize, list2cmdline)
 from ._version import __version__
 from .errors import AdbError, AdbInstallError
-from ._adb import AdbConnection
-from PIL import Image, UnidentifiedImageError
 
 _DISPLAY_RE = re.compile(
     r'.*DisplayViewport{.*?valid=true, .*?orientation=(?P<orientation>\d+), .*?deviceWidth=(?P<width>\d+), deviceHeight=(?P<height>\d+).*'
@@ -41,21 +44,49 @@ _DISPLAY_RE = re.compile(
 
 class BaseDevice:
     """ Basic operation for a device """
-    def __init__(self, client: BaseClient, serial: str):
+    def __init__(self, client: BaseClient, serial: str = None, transport_id: int = None):
         self._client = client
         self._serial = serial
+        self._transport_id: int = transport_id
         self._properties = {}  # store properties data
 
+        if not serial and not transport_id:
+            raise AdbError("serial, transport_id must set atleast one")
+
     @property
     def serial(self) -> str:
         return self._serial
 
-    def _get_with_command(self, cmd: str) -> str:
-        with self._client._connect() as c:
-            cmds = ["host-serial", self._serial, cmd]
-            c.send_command(":".join(cmds))
+    def open_transport(self, command: str = None, timeout: float = None) -> AdbConnection:
+        # connect has it own timeout
+        c = self._client._connect()
+        if timeout:
+            c.conn.settimeout(timeout)
+
+        if command:
+            if self._transport_id:
+                c.send_command(f"host-transport-id:{self._transport_id}:{command}")
+            elif self._serial:
+                c.send_command(f"host-serial:{self._serial}:{command}")
+            else:
+                raise RuntimeError
             c.check_okay()
-            return c.read_string_block()
+        else:
+            if self._transport_id:
+                c.send_command(f"host:transport-id:{self._transport_id}")
+            elif self._serial:
+                # host:tport:serial:xxx is also fine, but receive 12 bytes
+                # recv: 4f 4b 41 59 14 00 00 00 00 00 00 00              OKAY........
+                # so here use host:transport
+                c.send_command(f"host:transport:{self._serial}")
+            else:
+                raise RuntimeError
+            c.check_okay()
+        return c
+
+    def _get_with_command(self, cmd: str) -> str:
+        c = self.open_transport(cmd)
+        return c.read_string_block()
 
     def get_state(self) -> str:
         """ return device state {offline,bootloader,device} """
@@ -69,6 +100,21 @@ class BaseDevice:
         """ example return: usb:12345678Y """
         return self._get_with_command("get-devpath")
 
+    def get_features(self) -> str:
+        """
+        Return example:
+            'abb_exec,fixed_push_symlink_timestamp,abb,stat_v2,apex,shell_v2,fixed_push_mkdir,cmd'
+        """
+        return self._get_with_command("features")
+
+    @property
+    def info(self) -> dict:
+        return {
+            "serialno": self.get_serialno(),
+            "devpath": self.get_devpath(),
+            "state": self.get_state(),
+        }
+
     def __repr__(self):
         return "AdbDevice(serial={})".format(self.serial)
 
@@ -108,7 +154,7 @@ class BaseDevice:
               cmdargs: Union[str, list, tuple],
               stream: bool = False,
               timeout: Optional[float] = None,
-              rstrip=True) -> str:
+              rstrip=True) -> typing.Union[AdbConnection, str]:
         """Run shell inside device and get it's content
 
         Args:
@@ -117,55 +163,119 @@ class BaseDevice:
             timeout (float): set shell timeout
 
         Returns:
-            string of output
+            string of output when stream is False
+            AdbConnection when stream is True
 
         Raises:
+            AdbTimeout
 
         Examples:
             shell("ls -l")
             shell(["ls", "-l"])
             shell("ls | grep data")
         """
-        ret = self._client.shell(self._serial, cmdargs, stream=stream, timeout=timeout)
+        if isinstance(cmdargs, (list, tuple)):
+            cmdargs = list2cmdline(cmdargs)
         if stream:
-            return ret
-        return ret.rstrip() if rstrip else ret
+            timeout = None
+        c = self.open_transport(timeout=timeout)
+        c.send_command("shell:" + cmdargs)
+        c.check_okay()
+        if stream:
+            return c
+        output = c.read_until_close()
+        return output.rstrip() if rstrip else output
 
-    @deprecated(deprecated_in="0.2.4",
-                removed_in="0.3.0",
-                current_version=__version__,
-                details="use shell function instead, eg shell(\"ls -l\")")
-    def shell_output(self, *args) -> str:
-        return self._client.shell(self._serial, args)
+    def shell2(self,
+              cmdargs: Union[str, list, tuple],
+              timeout: Optional[float] = None,
+              rstrip=True) -> ShellReturn:
+        """
+        Run shell command with detail output
 
-    def forward(self, local: str, remote: str):
-        return self._client.forward(self._serial, local, remote)
+        Returns:
+            ShellOutput
+        
+        Raises:
+            AdbTimeout
+        """
+        if isinstance(cmdargs, (list, tuple)):
+            cmdargs = list2cmdline(cmdargs)
+        assert isinstance(cmdargs, str)
+        MAGIC = "X4EXIT:"
+        newcmd = cmdargs + f"; echo {MAGIC}$?"
+        output = self.shell(newcmd, timeout=timeout, rstrip=rstrip)
+        rindex = output.rfind(MAGIC)
+        if rindex == -1:  # normally will not possible
+            raise AdbError("shell output invalid", output)
+        returncoode = int(output[rindex + len(MAGIC):])
+        return ShellReturn(command=cmdargs,
+                           returncode=returncoode,
+                           output=output[:rindex])
+
+    def forward(self, local: str, remote: str, norebind: bool = False):
+        args = ["forward"]
+        if norebind:
+            args.append("norebind")
+        args.append(local+";" + remote)
+        self.open_transport(":".join(args))
 
     def forward_port(self, remote: Union[int, str]) -> int:
         """ forward remote port to local random port """
         if isinstance(remote, int):
             remote = "tcp:" + str(remote)
         for f in self.forward_list():
-            if f.serial == self._serial and f.remote == remote and f.local.startswith(
-                    "tcp:"):
+            if f.serial == self._serial and f.remote == remote and f.local.startswith("tcp:"):
                 return int(f.local[len("tcp:"):])
         local_port = get_free_port()
-        self._client.forward(self._serial, "tcp:" + str(local_port), remote)
+        self.forward("tcp:" + str(local_port), remote)
         return local_port
 
     def forward_list(self):
+        c = self.open_transport("list-forward")
+        content = c.read_string_block()
+        for line in content.splitlines():
+            parts = line.split()
+            if len(parts) != 3:
+                continue
+            yield ForwardItem(*parts)
         return self._client.forward_list(self._serial)
 
-    def reverse(self, remote: str, local: str):
-        return self._client.reverse(self._serial, remote, local)
+    def reverse(self, remote: str, local: str, norebind: bool = False):
+        """
+        Args:
+            serial (str): device serial
+            remote, local (str):
+                - tcp:<port>
+                - localabstract:<unix domain socket name>
+                - localreserved:<unix domain socket name>
+                - localfilesystem:<unix domain socket name>
+            norebind (bool): fail if already reversed when set to true
+
+        Raises:
+            AdbError
+        """
+        args = ["forward"]
+        if norebind:
+            args.append("norebind")
+        args.append(local+";" + remote)
+        self.open_transport(":".join(args))
 
     def reverse_list(self):
-        return self._client.reverse_list(self._serial)
+        c = self.open_transport()
+        c.send_command("reverse:list-forward")
+        c.check_okay()
+        content = c.read_string_block()
+        for line in content.splitlines():
+            parts = line.split()
+            if len(parts) != 3:
+                continue
+            yield ReverseItem(*parts[1:])
 
     def push(self, local: str, remote: str):
         self.adb_output("push", local, remote)
 
-    def create_connection(self, network: Network, address: Union[int, str]):
+    def create_connection(self, network: Network, address: Union[int, str]) -> socket.socket:
         """
         Used to connect a socket (unix of tcp) on the device
 
@@ -175,9 +285,7 @@ class BaseDevice:
         Raises:
             AssertionError, ValueError
         """
-        c = self._client._connect()
-        c.send_command("host:transport:" + self._serial)
-        c.check_okay()
+        c = self.open_transport()
         if network == Network.TCP:
             assert isinstance(address, int)
             c.send_command("tcp:" + str(address))
@@ -191,8 +299,32 @@ class BaseDevice:
             c.check_okay()
         else:
             raise ValueError("Unsupported network type", network)
+        c._finalizer.detach()
         return c.conn
 
+    def root(self):
+        """ restart adbd as root
+        
+        Return example:
+            cannot run as root in production builds
+        """
+        # Ref: https://github.com/Swind/pure-python-adb/blob/master/ppadb/command/transport/__init__.py#L179
+        c = self.open_transport()
+        c.send_command("root:")
+        c.check_okay()
+        return c.read_string_block()
+
+    def tcpip(self, port: int):
+        """ restart adbd listening on TCP on PORT
+
+        Return example:
+            restarting in TCP mode port: 5555
+        """
+        c = self.open_transport()
+        c.send_command("tcpip:" + str(port))
+        c.check_okay()
+        return c.read_until_close()
+
 
 class Property():
     def __init__(self, d: BaseDevice):
@@ -248,6 +380,10 @@ class Sync():
         finally:
             c.close()
 
+    def exists(self, path: str) -> bool:
+        finfo = self.stat(path)
+        return finfo.mtime is not None
+
     def stat(self, path: str) -> FileInfo:
         with self._prepare_sync(path, "STAT") as c:
             assert "STAT" == c.read_string(4)
@@ -276,7 +412,7 @@ class Sync():
 
     def push(self,
              src: typing.Union[pathlib.Path, str, bytes, bytearray, typing.BinaryIO],
-             dst: str,
+             dst: typing.Union[pathlib.Path, str],
              mode: int = 0o755,
              check: bool = False) -> int:
         # IFREG: File Regular
@@ -290,6 +426,9 @@ class Sync():
         else:
             if not hasattr(src, "read"):
                 raise TypeError("Invalid src type: %s" % type(src))
+
+        if isinstance(dst, pathlib.Path):
+            dst = dst.as_posix()
         path = dst + "," + str(stat.S_IFREG | mode)
         total_size = 0
         with self._prepare_sync(path, "SEND") as c:
@@ -298,7 +437,6 @@ class Sync():
                 while True:
                     chunk = r.read(4096)
                     if not chunk:
-                        print("d:", datetime)
                         mtime = int(datetime.datetime.now().timestamp())
                         c.conn.send(b"DONE" + struct.pack("<I", mtime))
                         break
@@ -322,7 +460,7 @@ class Sync():
                 if cmd == _FAIL:
                     str_size = struct.unpack("<I", c.read(4))[0]
                     error_message = c.read_string(str_size)
-                    raise AdbError(error_message)
+                    raise AdbError(error_message, path)
                 elif cmd == _DONE:
                     break
                 elif cmd == _DATA:
@@ -333,7 +471,7 @@ class Sync():
                     yield chunk
                 else:
                     raise AdbError("Invalid sync cmd", cmd)
-    
+
     def read_bytes(self, path: str) -> bytes:
         return b''.join(self.iter_content(path))
 
@@ -378,7 +516,7 @@ class AdbDevice(BaseDevice):
         except UnidentifiedImageError:
             w, h = self.window_size()
             return Image.new("RGB", (w, h), (220, 120, 100))
-            
+
     def switch_screen(self, status: bool):
         """
         turn screen on/off
@@ -520,19 +658,19 @@ class AdbDevice(BaseDevice):
         m = re.search(r'inet\s*addr:(.*?)\s', result, re.DOTALL)
         if m:
             return m.group(1)
-        
+
         # Huawei P30, has no ifconfig
         result = self.shell(['ip', 'addr', 'show', 'dev', 'wlan0'])
         m = re.search(r'inet (\d+.*?)/\d+', result)
         if m:
             return m.group(1)
-        
+
         # On VirtualDevice, might use eth0
         result = self.shell(['ifconfig', 'eth0'])
         m = re.search(r'inet\s*addr:(.*?)\s', result, re.DOTALL)
         if m:
             return m.group(1)
-        
+
         raise AdbError("fail to parse wlan ip")
 
 
@@ -702,7 +840,7 @@ class AdbDevice(BaseDevice):
             None or dict(version_name, version_code, signature)
         """
         output = self.shell(['dumpsys', 'package', package_name])
-        m = re.compile(r'versionName=(?P<name>[\d.]+)').search(output)
+        m = re.compile(r'versionName=(?P<name>[\w.]+)').search(output)
         version_name = m.group('name') if m else ""
         m = re.compile(r'versionCode=(?P<code>\d+)').search(output)
         version_code = m.group('code') if m else ""
@@ -718,11 +856,11 @@ class AdbDevice(BaseDevice):
 
         time_regex = r"[-\d]+\s+[:\d]+"
         m = re.compile(f"firstInstallTime=({time_regex})").search(output)
-        first_install_time = datetime.strptime(
+        first_install_time = datetime.datetime.strptime(
             m.group(1), "%Y-%m-%d %H:%M:%S") if m else None
 
         m = re.compile(f"lastUpdateTime=({time_regex})").search(output)
-        last_update_time = datetime.strptime(
+        last_update_time = datetime.datetime.strptime(
             m.group(1).strip(), "%Y-%m-%d %H:%M:%S") if m else None
 
         return dict(package_name=package_name,
@@ -823,23 +961,41 @@ class AdbDevice(BaseDevice):
 
         Returns:
             content of xml
+        
+        Raises:
+            AdbError
         """
         output = self.shell(
             'uiautomator dump /data/local/tmp/uidump.xml && echo success')
         if "success" not in output:
-            raise RuntimeError("uiautomator dump failed")
+            raise AdbError("uiautomator dump failed", output)
 
         buf = b''
         for chunk in self.sync.iter_content("/data/local/tmp/uidump.xml"):
             buf += chunk
         return buf.decode("utf-8")
 
-    @retry(AdbError, delay=.5, tries=3, jitter=.1)
-    def current_app(self):
+    @deprecated(deprecated_in="0.15.0",
+                removed_in="1.0.0",
+                current_version=__version__,
+                details="Use app_current instead")
+    def current_app(self) -> dict:
         """
         Returns:
             dict(package, activity, pid?)
 
+        Raises:
+            AdbError
+        """
+        info = self.app_current()
+        return asdict(info)
+
+    @retry(AdbError, delay=.5, tries=3, jitter=.1)
+    def app_current(self) -> RunningAppInfo:
+        """
+        Returns:
+            RunningAppInfo(package, activity, pid?)  pid can be 0
+
         Raises:
             AdbError
         """
@@ -856,8 +1012,17 @@ class AdbDevice(BaseDevice):
         )
         m = _focusedRE.search(self.shell(['dumpsys', 'window', 'windows']))
         if m:
-            return dict(package=m.group('package'),
-                        activity=m.group('activity'))
+            return RunningAppInfo(package=m.group('package'),
+                               activity=m.group('activity'))
+
+        # search mResumedActivity
+        # https://stackoverflow.com/questions/13193592/adb-android-getting-the-name-of-the-current-activity
+        package = None
+        output = self.shell(['dumpsys', 'activity', 'activities'])
+        _recordRE = re.compile(r'mResumedActivity: ActivityRecord\{.*?\s+(?P<package>[^\s]+)/(?P<activity>[^\s]+)\s.*?\}')
+        m = _recordRE.search(output)
+        if m:
+            package = m.group("package")
 
         # try: adb shell dumpsys activity top
         _activityRE = re.compile(
@@ -867,9 +1032,12 @@ class AdbDevice(BaseDevice):
         ms = _activityRE.finditer(output)
         ret = None
         for m in ms:
-            ret = dict(package=m.group('package'),
-                       activity=m.group('activity'),
-                       pid=int(m.group('pid')))
+            ret = RunningAppInfo(package=m.group('package'),
+                                 activity=m.group('activity'),
+                                 pid=int(m.group('pid')))
+            if ret.package == package:
+                return ret
+
         if ret:  # get last result
             return ret
         raise AdbError("Couldn't get focused app")
@@ -877,7 +1045,7 @@ class AdbDevice(BaseDevice):
     def remove(self, path: str):
         """ rm device file """
         self.shell(["rm", path])
-    
+
 
     def screenrecord(self, remote_path=None, no_autostart=False):
         """
@@ -885,8 +1053,124 @@ class AdbDevice(BaseDevice):
             remote_path: device video path
             no_autostart: do not start screenrecord, when call this method
         """
+        if self.shell2("which screenrecord").returncode != 0:
+            raise AdbError("screenrecord command not found")
         return _ScreenRecord(self, remote_path, autostart=not no_autostart)
 
+    def _start_recording(self, h264_filename: str):
+        """ start video recording """
+        return self._scrcpy.start_recording(h264_filename)
+    
+    def _stop_recording(self):
+        """ stop video recording """
+        return self._scrcpy.stop_recording()
+    
+    def _is_recording(self) -> bool:
+        return self._scrcpy.running
+
+    @property
+    def _scrcpy(self) -> "ScrcpyClient":
+        return ScrcpyClient.get_instance(self)
+
+
+class ScrcpyClient:
+    """
+    # -y  overwrite output files
+    ffmpeg -i "output.h264" -c:v copy -f mp4 -y "video.mp4"
+
+    遗留问题：秒表视频，转化后的视频时长不对啊（本来5s的，转化后变成了20s）
+
+    https://stackoverflow.com/questions/21263064/how-to-wrap-h264-into-a-mp4-container
+    """
+    __instances = {}
+
+    def __init__(self, d: AdbDevice, h264_filename: str = None):
+        self._d = d
+        self._filename = h264_filename
+        self._conn: AdbConnection = None
+        self._done_event = threading.Event()
+
+    @classmethod
+    def get_instance(cls, d: AdbDevice) -> "ScrcpyClient":
+        if d in cls.__instances:
+            return cls.__instances[d]
+        c = cls.__instances[d] = ScrcpyClient(d)
+        return c
+    
+    @property
+    def running(self) -> bool:
+        return self._conn and not self._conn.closed
+    
+    def start_recording(self, h264_filename: str):
+        if self.running:
+            raise RuntimeError("already running")
+        self._filename = h264_filename
+        self._start()
+
+    def stop_recording(self):
+        if not self.running:
+            raise RuntimeError("already stopped")
+        self._stop()
+
+    def _start(self):
+        curdir = pathlib.Path(__file__).absolute().parent
+        device_jar_path = "/data/local/tmp/scrcpy-server.jar"
+        scrcpy_server_jar_path = curdir.joinpath("binaries/scrcpy-server-1.24.jar")
+
+        self._d.sync.push(scrcpy_server_jar_path, device_jar_path)
+
+        opts = [
+            'control=false', 'bit_rate=8000000', 'tunnel_forward=true',
+            'lock_video_orientation=-1', 'send_dummy_byte=false',
+            "send_device_meta=false", "send_frame_meta=true", "downsize_on_error=true"
+        ]
+        cmd = [
+            'CLASSPATH=' + device_jar_path, 'app_process', '/',
+            '--nice-name=scrcpy-server', 'com.genymobile.scrcpy.Server', '1.24'
+        ] + opts
+        _c = self._d.shell(cmd, stream=True)
+        c: AdbConnection = _c
+        del(_c)
+        message = c.conn.recv(100).decode('utf-8')
+        print("Scrcpy:", message)
+        self._conn = c
+        threading.Thread(name="scrcpy_main", target=self._copy2null, args=(c.conn,), daemon=True).start()
+        time.sleep(0.1)
+        stream_sock = self._safe_dial_scrcpy()
+        fh = pathlib.Path(self._filename).open("wb")
+        threading.Thread(name="socket_copy", target=self._copy2file, args=(stream_sock, fh), daemon=True).start()
+
+    @retry(AdbError, tries=10, delay=0.1, jitter=0.01)
+    def _safe_dial_scrcpy(self) -> socket.socket:
+        return self._d.create_connection(Network.LOCAL_ABSTRACT, "scrcpy")
+
+    def _copy2null(self, s: socket.socket):
+        while True:
+            try:
+                chunk = s.recv(1024)
+                if chunk == b"":
+                    print("O:", chunk.decode('utf-8'))
+                    break
+            except:
+                break
+        print("Scrcpy mainThread stopped")
+
+    def _copy2file(self, s: socket.socket, fh: typing.BinaryIO):
+        while True:
+            chunk = s.recv(1<<16)
+            if not chunk:
+                break
+            fh.write(chunk)
+        fh.close()
+        print("Copy h264 stream finished", flush=True)
+        self._done_event.set()
+    
+    def _stop(self) -> bool:
+        self._conn.close()
+        self._done_event.wait(timeout=3.0)
+        time.sleep(1)
+        self._done_event.clear()
+
 
 class _ScreenRecord():
     def __init__(self, d: AdbDevice, remote_path=None, autostart=False):
@@ -907,7 +1191,18 @@ class _ScreenRecord():
         if self._started:
             warnings.warn("screenrecord already started", UserWarning)
             return
-        self._stream: AdbConnection = self._d.shell(["screenrecord", self._remote_path],
+        # end first line with \ to avoid the empty line!
+        script_content = textwrap.dedent("""\
+        #!/system/bin/sh
+        # generate by adbutils
+        screenrecord "$1" &
+        PID=$!
+        read ANY
+        kill -INT $PID
+        wait
+        """).encode('utf-8')
+        self._d.sync.push(script_content, "/sdcard/adbutils-screenrecord.sh")
+        self._stream: AdbConnection = self._d.shell(["sh", "/sdcard/adbutils-screenrecord.sh", self._remote_path],
                                      stream=True)
         self._started = True
 
@@ -918,19 +1213,15 @@ class _ScreenRecord():
 
         if self._stopped:
             return
-        self._stream.send(b"\003")
+        self._stream.send(b"\n")
         self._stream.read_until_close()
         self._stream.close()
         self._stopped = True
 
-    def stop_and_pull(self, path: str):
+    def stop_and_pull(self, path: typing.Union[str, pathlib.Path]):
         """ pull remote to local and remove remote file """
+        if isinstance(path, pathlib.Path):
+            path = path.as_posix()
         self.stop()
         self._d.sync.pull(self._remote_path, path)
-        self._d.remove(self._remote_path)
-
-    def close(self):  # alias of stop
-        return self.stop()
-
-    def close_and_pull(self, path: str):  # alias of stop_and_pull
-        return self.stop_and_pull(path=path)
+        self._d.remove(self._remote_path)
\ No newline at end of file
