@@ -1,7 +1,7 @@
 # coding: utf-8
 
 import pytest
-from adbutils import adb
+from adbutils import adb, AdbDevice
 
 
 @pytest.fixture(scope="session")
@@ -11,6 +11,6 @@ def device():
 
 
 @pytest.fixture
-def device_tmp_path(device):
+def device_tmp_path(device: AdbDevice):
     yield "/data/local/tmp/hi.txt"
-    device.shell(["rm", "/data/local/tmp/hi.txt"])
+    device.remove("/data/local/tmp/hi.txt")
